{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# 39 - APIs"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Database Programming with Function Calls: SQL/CLI & JDBC\n",
    "\n",
    "This approach uses function calls, which is a dynamic approach to database programming.\n",
    "\n",
    "**SQL Call Level Interface (SQL/CLI)** is part of the SQL standard and is a standardization of ODBC.\n",
    "\n",
    "### SQL/CLI: Using C as the Host Language\n",
    "\n",
    "- **Environment Record**\n",
    "    - Keeps environment information\n",
    "    - Track one or more database connections\n",
    "- **Connection Record**\n",
    "    - Keeps track of information needed for a particular database connection\n",
    "- **Statement Record**\n",
    "    - Keeps track of the information needed for one SQL statement\n",
    "- **Description Record**\n",
    "    - Keeps track of information about tuples or parameters\n",
    "\n",
    "Here is query for single tuple as a result:\n",
    "\n",
    "```C\n",
    "#include sqlchi.h;\n",
    "void printSal() {\n",
    "\n",
    "    SQLHSTMT stmt1;\n",
    "    SQLHDBC con1;\n",
    "    SQLHENV env1;\n",
    "    SQLRETURN ret1, ret2, ret3, ret4;\n",
    "    ret1 = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &env1);\n",
    "    if (!ret1) ret2 = SQLAllocHandle(SQL_HANDLE_DBC, env1, &con1) else exit;\n",
    "    if (!ret2) ret3 = SQLConnect(con1, \"dbs\", SQL_NTS, \"js\", SQL_NTS, \"xyz\", SQL_NTS) else exit;\n",
    "    if (!ret3) ret4 = SQLAllocHandle(SQL_HANDLE_STMT, con1, &stmt1) else exit;\n",
    "\n",
    "    SQLPrepare(stmt1, \"select Lname, Salary from EMPLOYEE where Ssn = ?\", SQL_NTS);\n",
    "\n",
    "    prompt(\"Enter a Social Security Number: \", ssn);\n",
    "    SQLBindParameter(stmt1, 1, SQL_CHAR, &ssn, 9, &fetchlen1);\n",
    "    ret1 = SQLExecute(stmt1);\n",
    "    if (!ret1) {\n",
    "        SQLBindCol(stmt1, 1, SQL_CHAR, &lname, 15, &fetchlen1);\n",
    "        SQLBindCol(stmt1, 2, SQL_FLOAT, &salary, 4, &fetchlen2);\n",
    "        ret2 = SQLFetch(stmt1);\n",
    "        if (!ret2) {\n",
    "            printf(ssn, lname, salary);\n",
    "        } else {\n",
    "            printf(\"Social Security Number does not exist: \", ssn);\n",
    "        }\n",
    "    }\n",
    "}\n",
    "```\n",
    "\n",
    "Here is a query for a collection of tuples as a result:\n",
    "\n",
    "```C\n",
    "#include sqlchi.h;\n",
    "void printDepartmentEmps() {\n",
    "\n",
    "    // Unchanged from previous example\n",
    "    SQLHSTMT stmt1;\n",
    "    SQLHDBC con1;\n",
    "    SQLHENV env1;\n",
    "    SQLRETURN ret1, ret2, ret3, ret4;\n",
    "    ret1 = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &env1);\n",
    "    if (!ret1) ret2 = SQLAllocHandle(SQL_HANDLE_DBC, env1, &con1) else exit;\n",
    "    if (!ret2) ret3 = SQLConnect(con1, \"dbs\", SQL_NTS, \"js\", SQL_NTS, \"xyz\", SQL_NTS) else exit;\n",
    "    if (!ret3) ret4 = SQLAllocHandle(SQL_HANDLE_STMT, con1, &stmt1) else exit;\n",
    "\n",
    "    SQLPrepare(stmt1, \"select Lname, Salary from EMPLOYEE where Dno = ?\", SQL_NTS);\n",
    "\n",
    "    prompt(\"Enter the Department Number: \", ssn);\n",
    "    SQLBindParameter(stmt1, 1, SQL_INTEGER, &dno, 4, &fetchlen1);\n",
    "    ret1 = SQLExecute(stmt1);\n",
    "    if (!ret1) {\n",
    "        SQLBindCol(stmt1, 1, SQL_CHAR, &lname, 15, &fetchlen1);\n",
    "        SQLBindCol(stmt1, 2, SQL_FLOAT, &salary, 4, &fetchlen2);\n",
    "        ret2 = SQLFetch(stmt1);\n",
    "        while (!ret2) {\n",
    "            printf(lname, salary);\n",
    "            ret2 = SQLFETCH(stmt1);\n",
    "        }\n",
    "    }\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### JDBC: SQL Function Calls for Java Programming\n",
    "\n",
    "**JDBC** uses function libraries for Java. A single Java program can connect to several databases, which are called data sources by the Java program.\n",
    "\n",
    "A **JDBC Driver** must be loaded explicitly:\n",
    "\n",
    "```Java\n",
    "Class.forName(\"oracle.jdbc.driver.OracleDriver\");\n",
    "```\n",
    "\n",
    "JDBC utilizes a `Connection` object, `Statement` object (with `PreparedStatement` and `CallableStatement` subclasses), the question mark (`?`) symbol (represents a statement parameter, determined at runtime), and a `ResultSet` object (holds results of the query).\n",
    "\n",
    "Here is a query for a single tuple as a result:\n",
    "\n",
    "```Java\n",
    "import java.io.*;\n",
    "import java.sql.*;\n",
    "\n",
    "class getEmpInfo {\n",
    "    public static void main(String args[]) throws SQLException, IOException {\n",
    "\n",
    "        try {\n",
    "            Class.forName(\"oracle.jdbc.driver.OracleDriver\");\n",
    "        } catch (ClassNotFoundException x) {\n",
    "            System.out.println(\"Driver could not be loaded\");\n",
    "        }\n",
    "\n",
    "        String dbacct, passwrd, ssn, lname;\n",
    "        Double salary;\n",
    "        dbacct = readEntry(\"Enter database account:\");\n",
    "        passwrd = readEntry(\"Enter password:\");\n",
    "        Connection conn = DriverManager.getConnection(\"jdbc:oracle:oci8:\" + dbacct + \"/\" + passwrd);\n",
    "        String stmt1 = \"select Lname, Salary from EMPLOYEE where Ssn = ?\";\n",
    "        PreparedStatement p = conn.prepareStatement(stmt1);\n",
    "        ssn = readentry(\"Enter a Social Security Number:\");\n",
    "        p.clearParameters();\n",
    "        p.setString(1, ssn);\n",
    "        ResultSet r = p.executeQuery();\n",
    "        while (r.next()) {\n",
    "            lname = r.getString(1);\n",
    "            salary = r.getDouble(2);\n",
    "            System.out.println(lname + salary);\n",
    "        }\n",
    "    }\n",
    "}\n",
    "```\n",
    "\n",
    "Here is a query for multiple tuples:\n",
    "\n",
    "```Java\n",
    "import java.io.*;\n",
    "import java.sql.*;\n",
    "\n",
    "class getDepartmentEmps {\n",
    "    public static void main(String args[]) throws SQLException, IOException {\n",
    "\n",
    "        try {\n",
    "            Class.forName(\"oracle.jdbc.driver.OracleDriver\");\n",
    "        } catch (ClassNotFoundException x) {\n",
    "            System.out.println(\"Driver could not be loaded\");\n",
    "        }\n",
    "\n",
    "        String dbacct, passwrd, ssn, lname;\n",
    "        Double salary;\n",
    "        Integer dno;\n",
    "        dbacct = readEntry(\"Enter database account:\");\n",
    "        passwrd = readEntry(\"Enter password:\");\n",
    "        Connection conn = DriverManager.getConnection(\"jdbc:oracle:oci8:\" + dbacct + \"/\" + passwrd);\n",
    "\n",
    "        dno = readEntry(\"Enter a Department Number:\");\n",
    "        String q = \"select Lname, Salary from EMPLOYEE where Dno = \" + dno.tostring();\n",
    "        Statement s = conn.createStatement();\n",
    "        ResultSet r = s.executeQuery(q);\n",
    "        while (r.next()) {\n",
    "            lname = r.getString(1);\n",
    "            salary = r.getDouble(2);\n",
    "            System.out.println(lname + salary);\n",
    "        }\n",
    "    }\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Database Stored Procedures and SQL/PSM\n",
    "\n",
    "**Stored procedures** are program modules stored by the DBMS at the database server. These can be functions or procedures.\n",
    "\n",
    "```SQL\n",
    "CREATE PROCEDURE <procedure name> (<parameters>)\n",
    "<local declarations>\n",
    "<procedure body>;\n",
    "```\n",
    "\n",
    "**SQL/Persistent stored modules (SQL/PSM)** are extensions to SQL that include general-purpose programming constructs in SQL.\n",
    "\n",
    "**Persistent stored modules** are stored persistently by the DBMS. They are useful when a database program is needed by several applications, for reducing data transfer and communication cost between client and server in certain situations, and for enhancing modeling power provided by views.\n",
    "\n",
    "```SQL\n",
    "CREATE FUNCTION <function name> (<parameters>)\n",
    "RETURNS <return type>\n",
    "<local declarations>\n",
    "<function body>;\n",
    "```\n",
    "\n",
    "Each parameter has a **parameter type**, which is one of the SQL data types, and a **parameter mode**, which is `IN`, `OUT`, or `INOUT`.\n",
    "\n",
    "To call a stored procedure:\n",
    "\n",
    "```SQL\n",
    "CALL <procedure or function name> (<argument list>);\n",
    "```\n",
    "\n",
    "Here is a conditional branching statement:\n",
    "\n",
    "```SQL\n",
    "IF <condition> THEN <statement list>\n",
    "ELSEIF <condition> THEN <statement list>\n",
    "ELSEIF <condition> THEN <statement list>\n",
    "ELSE <statement list>\n",
    "END IF;\n",
    "```\n",
    "\n",
    "Here is a while loop:\n",
    "\n",
    "```SQL\n",
    "WHILE <condition> DO\n",
    "    <statement list>\n",
    "END WHILE;\n",
    "```\n",
    "```SQL\n",
    "REPEAT\n",
    "    <statement list>\n",
    "UNTIL <condition>\n",
    "END REPEAT;\n",
    "```\n",
    "\n",
    "And here is a for loop:\n",
    "\n",
    "```SQL\n",
    "FOR <loop name> AS <cursor name> CURSOR FOR <query> DO\n",
    "    <statement list>\n",
    "END FOR;\n",
    "```\n",
    "\n",
    "And here is how to declare a function:\n",
    "\n",
    "```SQL\n",
    "CREATE FUNCTION Dept_size(IN deptno INTEGER)\n",
    "RETURNS VARCHAR [7]\n",
    "DECLARE No_of_emps INTEGER;\n",
    "SELECT COUNT(*) INTO No_of_emps\n",
    "FROM EMPLOYEE WHERE Dno = deptno;\n",
    "IF No_of_emps > 100 THEN RETURN \"HUGE\"\n",
    "ELSEIF No_of_emps > 25 THEN RETURN \"LARGE\"\n",
    "ELSEIF No_of_emps > 10 THEN RETURN \"MEDIUM\"\n",
    "ELSE RETURN \"SMALL\"\n",
    "END IF;\n",
    "```"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
